MODULE dynnxt
   !!======================================================================
   !!                       ***  MODULE  dynnxt  ***
   !! Ocean dynamics: time stepping
   !!======================================================================
   !!======================================================================
   !! History :  OPA  !  1987-02  (P. Andrich, D. L Hostis)  Original code
   !!                 !  1990-10  (C. Levy, G. Madec)
   !!            7.0  !  1993-03  (M. Guyon)  symetrical conditions
   !!            8.0  !  1997-02  (G. Madec & M. Imbard)  opa, release 8.0
   !!            8.2  !  1997-04  (A. Weaver)  Euler forward step
   !!             -   !  1997-06  (G. Madec)  lateral boudary cond., lbc routine
   !!    NEMO    1.0  !  2002-08  (G. Madec)  F90: Free form and module
   !!             -   !  2002-10  (C. Talandier, A-M. Treguier) Open boundary cond.
   !!            2.0  !  2005-11  (V. Garnier) Surface pressure gradient organization
   !!            2.3  !  2007-07  (D. Storkey) Calls to BDY routines. 
   !!            3.1  !  2009-02  (G. Madec)  re-introduce the vvl option
   !!----------------------------------------------------------------------
  
   !!----------------------------------------------------------------------
   !!   dyn_nxt      : update the horizontal velocity from the momentum trend
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! ocean dynamics and tracers
   USE dom_oce         ! ocean space and time domain
   USE in_out_manager  ! I/O manager
   USE obc_oce         ! ocean open boundary conditions
   USE obcdyn          ! open boundary condition for momentum (obc_dyn routine)
   USE obcdyn_bt       ! 2D open boundary condition for momentum (obc_dyn_bt routine)
   USE obcvol          ! ocean open boundary condition (obc_vol routines)
   USE bdy_oce         ! unstructured open boundary conditions
   USE bdydta          ! unstructured open boundary conditions
   USE bdydyn          ! unstructured open boundary conditions
   USE dynspg_oce      ! type of surface pressure gradient
   USE lbclnk          ! lateral boundary condition (or mpp link)
   USE prtctl          ! Print control
   USE agrif_opa_update
   USE agrif_opa_interp
   USE domvvl          ! variable volume

   IMPLICIT NONE
   PRIVATE

   !! * Accessibility
   PUBLIC dyn_nxt                ! routine called by step.F90
   !! * Substitutions
#  include "domzgr_substitute.h90"
   !!----------------------------------------------------------------------
   !!  OPA 9.0 , LOCEAN-IPSL (2005) 
   !! $Id$ 
   !! This software is governed by the CeCILL licence see modipsl/doc/NEMO_CeCILL.txt 
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE dyn_nxt ( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dyn_nxt  ***
      !!                   
      !! ** Purpose :   Compute the after horizontal velocity from the 
      !!      momentum trend.
      !!
      !! ** Method  :   Apply lateral boundary conditions on the trends (ua,va) 
      !!      through calls to routine lbc_lnk.
      !!      After velocity is compute using a leap-frog scheme environment:
      !!         (ua,va) = (ub,vb) + 2 rdt (ua,va)
      !!      Note that if lk_dynspg_flt=T, the time stepping has already been
      !!      performed in dynspg module
      !!      Time filter applied on now horizontal velocity to avoid the
      !!      divergence of two consecutive time-steps and swap of dynamics
      !!      arrays to start the next time step:
      !!         (ub,vb) = (un,vn) + atfp [ (ub,vb) + (ua,va) - 2 (un,vn) ]
      !!         (un,vn) = (ua,va) 
      !!
      !! ** Action : - Update ub,vb arrays, the before horizontal velocity
      !!             - Update un,vn arrays, the now horizontal velocity
      !!
      !!----------------------------------------------------------------------
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index
      !!
      INTEGER  ::   jk   ! dummy loop indices
      REAL(wp) ::   z2dt         ! temporary scalar
      !!----------------------------------------------------------------------

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dyn_nxt : time stepping'
         IF(lwp) WRITE(numout,*) '~~~~~~~'
      ENDIF

      ! Local constant initialization
      z2dt = 2. * rdt
      IF( neuler == 0 .AND. kt == nit000 )  z2dt = rdt

      !! Explicit physics with thickness weighted updates

      ! Lateral boundary conditions on ( ua, va )
      CALL lbc_lnk( ua, 'U', -1. )
      CALL lbc_lnk( va, 'V', -1. )

      ! Next velocity
      ! -------------
#if defined key_dynspg_flt
      ! Leap-frog time stepping already done in dynspg_flt.F routine
#else
      IF( lk_vvl ) THEN                                  ! Varying levels
         !RB_vvl scale factors are wrong at this point
         DO jk = 1, jpkm1
            ua(ji,jj,jk) = (        ub(:,:,jk) * fse3u(:,:,jk)     &
               &           + z2dt * ua(:,:,jk) * fse3u(:,:,jk) )   &
               &         / fse3u(:,:,jk) * umask(:,:,jk)
            va(ji,jj,jk) = (        vb(:,:,jk) * fse3v(:,:,jk)     &
               &           + z2dt * va(:,:,jk) * fse3v(:,:,jk) )   &
               &         / fse3v(:,:,jk) * vmask(:,:,jk)
         END DO
      ELSE
         DO jk = 1, jpkm1
                  ! Leap-frog time stepping
            ua(:,:,jk) = ( ub(:,:,jk) + z2dt * ua(:,:,jk) ) * umask(:,:,jk)
            va(:,:,jk) = ( vb(:,:,jk) + z2dt * va(:,:,jk) ) * vmask(:,:,jk)
         END DO
      ENDIF

# if defined key_obc
      ! Update (ua,va) along open boundaries (only in the rigid-lid case)
      CALL obc_dyn( kt )

      IF ( lk_dynspg_exp .OR. lk_dynspg_ts ) THEN
         !Flather boundary condition :
         !        - Update sea surface height on each open boundary
         !                 sshn (= after ssh) for explicit case
         !                 sshn_b (= after ssha_b) for time-splitting case
         !        - Correct the barotropic velocities
         CALL obc_dyn_bt( kt )

         !Boundary conditions on sshn ( after ssh)
         CALL lbc_lnk( sshn, 'T', 1. )

         IF(ln_ctl) THEN         ! print sum trends (used for debugging)
            CALL prt_ctl(tab2d_1=sshn, clinfo1=' ssh      : ', mask1=tmask)
         ENDIF

         IF ( ln_vol_cst ) CALL obc_vol( kt )

      ENDIF

# elif defined key_bdy 
      ! Update (ua,va) along open boundaries (for exp or ts options).
      IF ( lk_dynspg_exp .or. lk_dynspg_ts ) THEN
  
         CALL bdy_dyn_frs( kt )

         IF ( ln_bdy_fla ) THEN

            ua_e(:,:)=0.0
            va_e(:,:)=0.0

            ! Set these variables for use in bdy_dyn_fla
            hu_e(:,:) = hu(:,:)
            hv_e(:,:) = hv(:,:)

            DO jk = 1, jpkm1
               !! Vertically integrated momentum trends
               ua_e(:,:) = ua_e(:,:) + fse3u(:,:,jk) * umask(:,:,jk) * ua(:,:,jk)
               va_e(:,:) = va_e(:,:) + fse3v(:,:,jk) * vmask(:,:,jk) * va(:,:,jk)
            END DO

            DO jk = 1 , jpkm1
               ua(:,:,jk) = ua(:,:,jk) - ua_e(:,:) * hur(:,:)
               va(:,:,jk) = va(:,:,jk) - va_e(:,:) * hvr(:,:)
            END DO

            CALL bdy_dta_bt( kt+1, 0)
            CALL bdy_dyn_fla

         ENDIF

         DO jk = 1 , jpkm1
            ua(:,:,jk) = ua(:,:,jk) + ua_e(:,:) * hur(:,:)
            va(:,:,jk) = va(:,:,jk) + va_e(:,:) * hvr(:,:)
         END DO

      ENDIF

# endif
# if defined key_agrif
      CALL Agrif_dyn( kt )
# endif
#endif

      ! Time filter and swap of dynamics arrays
      ! ------------------------------------------
      IF( neuler == 0 .AND. kt == nit000 ) THEN
         DO jk = 1, jpkm1                                 
            ub(:,:,jk) = un(:,:,jk) 
            vb(:,:,jk) = vn(:,:,jk)
            un(:,:,jk) = ua(:,:,jk)
            vn(:,:,jk) = va(:,:,jk)
         END DO
      ELSE
         IF( lk_vvl ) THEN                                ! Varying levels
            ! Not done
         ELSE                                             ! Fixed levels
!RB_vvl : should be done as in tranxt ?
            DO jk = 1, jpkm1                              ! filter applied on velocities
               ub(:,:,jk) = atfp * ( ub(:,:,jk) + ua(:,:,jk) ) + atfp1 * un(:,:,jk)
               vb(:,:,jk) = atfp * ( vb(:,:,jk) + va(:,:,jk) ) + atfp1 * vn(:,:,jk)
               un(:,:,jk) = ua(:,:,jk)
               vn(:,:,jk) = va(:,:,jk)
            END DO
         ENDIF
      ENDIF

      IF(ln_ctl)   THEN
         CALL prt_ctl(tab3d_1=un, clinfo1=' nxt  - Un: ', mask1=umask, &
            &         tab3d_2=vn, clinfo2=' Vn: ', mask2=vmask)
      ENDIF

#if defined key_agrif
      IF (.NOT.Agrif_Root())    CALL Agrif_Update_Dyn( kt )
#endif      

   END SUBROUTINE dyn_nxt

   !!======================================================================
END MODULE dynnxt
